<sly-select-tag
  :project-id="state.{{{widget.widget_id}}}.projectId"
  :tags.sync="state.{{{widget.widget_id}}}.tags"
  :options="data.{{{widget.widget_id}}}.options"
  :disabled="data.{{{widget.widget_id}}}.disabled"
></sly-select-tag>
