from supervisely.app.widgets.widget import Widget, generate_id
from supervisely.app.widgets.radio_table.radio_table import RadioTable
from supervisely.app.widgets.notification_box.notification_box import NotificationBox
from supervisely.app.widgets.done_label.done_label import DoneLabel
from supervisely.app.widgets.sly_tqdm.sly_tqdm import SlyTqdm, Progress
from supervisely.app.widgets.grid_gallery.grid_gallery import GridGallery
from supervisely.app.widgets.classic_table.classic_table import ClassicTable
from supervisely.app.widgets.confusion_matrix.confusion_matrix import ConfusionMatrix
from supervisely.app.widgets.project_selector.project_selector import ProjectSelector
from supervisely.app.widgets.element_button.element_button import ElementButton
from supervisely.app.widgets.button.button import Button
from supervisely.app.widgets.project_thumbnail.project_thumbnail import ProjectThumbnail
from supervisely.app.widgets.apexchart.apexchart import Apexchart
from supervisely.app.widgets.line_chart.line_chart import LineChart
from supervisely.app.widgets.heatmap_chart.heatmap_chart import HeatmapChart
from supervisely.app.widgets.table.table import Table
from supervisely.app.widgets.labeled_image.labeled_image import LabeledImage
from supervisely.app.widgets.text.text import Text
from supervisely.app.widgets.sidebar.sidebar import Sidebar
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.card.card import Card
from supervisely.app.widgets.select.select import Select
from supervisely.app.widgets.menu.menu import Menu
from supervisely.app.widgets.field.field import Field
from supervisely.app.widgets.input_number.input_number import InputNumber
from supervisely.app.widgets.video.video import Video
from supervisely.app.widgets.object_class_view.object_class_view import ObjectClassView
from supervisely.app.widgets.checkbox.checkbox import Checkbox
from supervisely.app.widgets.grid.grid import Grid
from supervisely.app.widgets.object_classes_list.object_classes_list import (
    ObjectClassesList,
)
from supervisely.app.widgets.empty.empty import Empty
from supervisely.app.widgets.one_of.one_of import OneOf
from supervisely.app.widgets.flexbox.flexbox import Flexbox
from supervisely.app.widgets.input.input import Input
from supervisely.app.widgets.select_team.select_team import SelectTeam
from supervisely.app.widgets.select_workspace.select_workspace import SelectWorkspace
from supervisely.app.widgets.select_project.select_project import SelectProject
from supervisely.app.widgets.select_dataset.select_dataset import SelectDataset
from supervisely.app.widgets.select_item.select_item import SelectItem
from supervisely.app.widgets.identity.identity import Identity
from supervisely.app.widgets.dataset_thumbnail.dataset_thumbnail import DatasetThumbnail
from supervisely.app.widgets.select_tag_meta.select_tag_meta import SelectTagMeta
from supervisely.app.widgets.video_thumbnail.video_thumbnail import VideoThumbnail
from supervisely.app.widgets.tabs.tabs import Tabs
