.beautiful-table {
  border-collapse: collapse;
}
.beautiful-table tr:nth-child(2n) {
  background-color: #f6f8fa;
}
.beautiful-table td,
.beautiful-table th {
  border: 1px solid #dfe2e5;
  padding: 6px 13px;
  text-align: center;
  line-height: 20px;
}
